@@ -1,12 +1,10 @@
 from ._biglist import Biglist, ListView, FileView, FileIterStat
-from ._serializer import Serializer
 
 
-__version__ = '0.5.6'
+__version__ = '0.6.2'
 
 
 __all__ = [
     'Biglist', 'ListView', 'FileView',
-    'Serializer',
     'FileIterStat',
 ]
