@@ -10,6 +10,7 @@ import queue
 import string
 import tempfile
 import threading
+import uuid
 from collections.abc import Sequence, Iterable
 from concurrent.futures import ThreadPoolExecutor, Future
 from contextlib import contextmanager
@@ -19,9 +20,9 @@ from typing import Iterator, Union, List, Dict, Optional, Tuple, Type, Callable
 from uuid import uuid4
 
 from upathlib import LocalUpath, Upath  # type: ignore
-from ._serializer import (
-    Serializer, PickleSerializer, CompressedPickleSerializer,
-    JsonSerializer, OrjsonSerializer, CompressedOrjsonSerializer,
+from upathlib.serializer import (
+    ByteSerializer, PickleSerializer, CompressedPickleSerializer,
+    JsonByteSerializer, OrjsonSerializer, CompressedOrjsonSerializer,
 )
 
 
@@ -50,6 +51,11 @@ class Dumper:
 
         self._task_file_data: Dict[Future, Tuple] = {}
 
+    def __del__(self):
+        if self._executor is not None:
+            self._executor.shutdown()
+            self._executor = None
+
     def _callback(self, t):
         self._sem.release()
         del self._task_file_data[t]
@@ -171,7 +177,7 @@ class Biglist(Sequence):
     @classmethod
     def register_storage_format(cls,
                                 name: str,
-                                serializer: Type[Serializer],
+                                serializer: Type[ByteSerializer],
                                 overwrite: bool = False,
                                 ):
         good = string.ascii_letters + string.digits + '-_'
@@ -187,9 +193,8 @@ class Biglist(Sequence):
         '''Subclass needs to customize this if it prefers to use
         a remote blobstore for temp Biglist.
         '''
-        path = tempfile.mkdtemp(dir=os.environ.get('TMPDIR', '/tmp'))
-        # This directory is already created by now.
-        path = LocalUpath(path)  # type: ignore
+        path = LocalUpath(os.path.abspath(tempfile.gettempdir()),
+                          str(uuid.uuid4()))  # type: ignore
         return path  # type: ignore
 
     @classmethod
@@ -278,8 +283,10 @@ class Biglist(Sequence):
                 path = LocalUpath(str(path.absolute()))
             if keep_files is None:
                 keep_files = True
-        if path.is_dir() and not path.is_empty_dir():
+        if path.is_dir():
             raise Exception(f'directory "{path}" already exists')
+        elif path.is_file():
+            raise FileExistsError(path)
 
         obj = cls(path, **kwargs)  # type: ignore
 
@@ -306,20 +313,7 @@ class Biglist(Sequence):
 
         return obj
 
-    def __init__(self,
-                 path: Union[str, Path, Upath],
-                 *,
-                 multi_writers: bool = True):
-        '''
-        `multi_writers`: if `true`, this Biglist is possibly being written to
-        (i.e. appended to) by other workers besides the current one.
-        The default value, `True`, is a conservative setting---it does no harm
-        other than slowing down random access. Setting it to `False` only when
-        you are sure the current object is the only one that may be writing
-        to the Biglist at this time.
-
-        TODO: find a better name for `multi_writers`.
-        '''
+    def __init__(self, path: Union[str, Path, Upath]):
         if isinstance(path, str):
             path = Path(path)
         if isinstance(path, Path):
@@ -328,7 +322,6 @@ class Biglist(Sequence):
         # `Upath` protocol.
         self.path = path
 
-        self._multi_writers = multi_writers
         self._data_files: Optional[list] = None
 
         self._read_buffer: Optional[list] = None
@@ -380,7 +373,7 @@ class Biglist(Sequence):
         if self.keep_files:
             self.flush()
         else:
-            self.destroy()
+            self.destroy(concurrency=0)
 
     def __getitem__(self, idx: int):  # type: ignore
         '''
@@ -402,7 +395,7 @@ class Biglist(Sequence):
             if n1 <= idx < n2:
                 return self._read_buffer[idx - n1]  # type: ignore
 
-        datafiles = self.get_data_files()
+        datafiles = self.get_data_files(lazy=True)
         length = sum(l for _, l in datafiles)
         idx = range(length + len(self._append_buffer))[idx]
 
@@ -440,33 +433,33 @@ class Biglist(Sequence):
         elif ndatafiles > 1:
             max_workers = min(3, ndatafiles)
             tasks = queue.Queue(max_workers)
-            executor = ThreadPoolExecutor(max_workers)
-            for i in range(max_workers):
-                t = executor.submit(
-                    self.load_data_file,
-                    self._data_dir / datafiles[i][0]
-                )
-                tasks.put(t)
-            nfiles_queued = max_workers
-
-            for _ in range(ndatafiles):
-                t = tasks.get()
-                data = t.result()
-
-                if nfiles_queued < ndatafiles:
+            with ThreadPoolExecutor(max_workers) as executor:
+                for i in range(max_workers):
                     t = executor.submit(
                         self.load_data_file,
-                        self._data_dir / datafiles[nfiles_queued][0]
+                        self._data_dir / datafiles[i][0]
                     )
                     tasks.put(t)
-                    nfiles_queued += 1
-                yield from data
+                nfiles_queued = max_workers
+
+                for _ in range(ndatafiles):
+                    t = tasks.get()
+                    data = t.result()
+
+                    if nfiles_queued < ndatafiles:
+                        t = executor.submit(
+                            self.load_data_file,
+                            self._data_dir / datafiles[nfiles_queued][0]
+                        )
+                        tasks.put(t)
+                        nfiles_queued += 1
+                    yield from data
 
         if self._append_buffer:
             yield from self._append_buffer
 
     def __len__(self) -> int:
-        z = self.get_data_files()
+        z = self.get_data_files(lazy=True)
         return sum(k for _, k in z) + len(self._append_buffer)
 
     def append(self, x) -> None:
@@ -490,7 +483,7 @@ class Biglist(Sequence):
             z.append((filename, length))
             self._data_info_file.write_json(z, overwrite=True)
 
-    def destroy(self) -> None:
+    def destroy(self, *, concurrency: int = None) -> None:
         '''
         Clears all the files and releases all in-memory data held by this object,
         so that the object is as if upon `__init__` with an empty directory pointed to
@@ -503,7 +496,7 @@ class Biglist(Sequence):
         self._read_buffer_file = None
         self._read_buffer_item_range = None
         self._append_buffer = []
-        self.path.rmrf()
+        self.path.rmrf(concurrency=concurrency)
 
     def extend(self, x: Iterable) -> None:
         for v in x:
@@ -519,7 +512,7 @@ class Biglist(Sequence):
 
     def file_view(self, file: Union[Upath, int]) -> FileView:
         if isinstance(file, int):
-            datafiles = self.get_data_files()
+            datafiles = self.get_data_files(lazy=True)
             file = self._data_dir / datafiles[file][0]
         return FileView(file, self.load_data_file)  # type: ignore
 
@@ -533,7 +526,7 @@ class Biglist(Sequence):
             for f, l in datafiles
         ]
 
-    def _flush(self):
+    def _flush(self, *, wait: bool = False):
         '''
         Persist the content of the in-memory buffer to a file,
         reset the buffer, and update relevant book-keeping variables.
@@ -552,16 +545,24 @@ class Biglist(Sequence):
             if not data_file.exists():
                 break
 
-        self._file_dumper.dump_file(
-            self.dump_data_file,
-            data_file,
-            self._append_buffer,
-        )
-        # This call will return quickly if the dumper has queue
-        # capacity for the file. The file meta data below
-        # will be updated as if the saving has completed, although
-        # it hasn't (it is only queued). This allows the waiting-to-be-saved
-        # data to be accessed property.
+        if wait:
+            self._file_dumper.wait()
+            self.dump_data_file(data_file, self._append_buffer)
+        else:
+            self._file_dumper.dump_file(
+                self.dump_data_file,
+                data_file,
+                self._append_buffer,
+            )
+            # This call will return quickly if the dumper has queue
+            # capacity for the file. The file meta data below
+            # will be updated as if the saving has completed, although
+            # it hasn't (it is only queued). This allows the waiting-to-be-saved
+            # data to be accessed property.
+
+        # TODO:
+        # what if dump fails later? the index file will be updated
+        # already assuming everything will be fine.
 
         self._append_buffer = []
         self._append_data_files_info(data_file.name, buffer_len)
@@ -579,11 +580,10 @@ class Biglist(Sequence):
         In summary, call this method once the user is done with adding elements
         to the list *in this session*, meaning in this run of the program.
         '''
-        self._flush()
-        self._file_dumper.wait()
+        self._flush(wait=True)
 
-    def get_data_files(self) -> list:
-        if not self._multi_writers and self._data_files is not None:
+    def get_data_files(self, lazy: bool = False) -> list:
+        if lazy and self._data_files is not None:
             return self._data_files
 
         if self._data_info_file.exists():
@@ -726,6 +726,6 @@ class ListView(Sequence):
         return self._list
 
 
-Biglist.register_storage_format('json', JsonSerializer)
+Biglist.register_storage_format('json', JsonByteSerializer)
 Biglist.register_storage_format('pickle-z', CompressedPickleSerializer)
 Biglist.register_storage_format('orjson-z', CompressedOrjsonSerializer)
